import multiprocessing
import datetime
import json
import os
import multiprocessing.managers
import random
import torch
from typing import Dict, List, Optional, Set, Tuple, Union, Any, TypedDict
from transformers import PreTrainedTokenizer, AutoTokenizer
from hf_ehr.config import Event, TokenizerConfigEntry, load_tokenizer_config_from_path, save_tokenizer_config_to_path, PATH_TO_FEMR_EXTRACT_v8, PATH_TO_TOKENIZER_CLMBR_v8_CONFIG, PATH_TO_TOKENIZER_DESC_v8_CONFIG, PATH_TO_TOKENIZER_COOKBOOK_v8_CONFIG, PATH_TO_TOKENIZER_CEHR_v8_CONFIG
from hf_ehr.data.datasets import FEMRDataset
from tqdm import tqdm
from omegaconf import OmegaConf, DictConfig

class TokenizerSeqLengthPerPatientCache(TypedDict):
    """Typing for `seq_length_per_patient.json` cache file"""
    timestamp: str
    tokenizer_metadata: Dict[str, Any]
    dataset_metadata: Dict[str, Any]
    seq_lengths: List[int]

def filter_tokenizer_config(tokenizer_config: List[TokenizerConfigEntry],
                            excluded_vocabs: Optional[Set[str]] = None,
                            min_code_occurrence_count: Optional[int] = None,
                            keep_n_max_occurrence_codes: Optional[int] = None,
                            **kwargs) -> Tuple[List[TokenizerConfigEntry], List[TokenizerConfigEntry]]:
    """
        Given a set of filters, applies them to the `tokenizer_config`. 
        Returns two lists -- one for valid tokens, one for invalid tokens.
    """
    valid_entries: List[TokenizerConfigEntry] = []
    invalid_entries: List[TokenizerConfigEntry] = []
    for entry in tokenizer_config:
        # Remove tokens from excluded vocabs
        if (
            excluded_vocabs is not None
            and entry.code.split("/")[0].lower() in excluded_vocabs
        ):
            invalid_entries.append(entry.to_token())
            continue
        # Remove tokens with < `min_code_occurrence_count` occurrences in our dataset
        if (
            min_code_occurrence_count is not None
            and entry.get_stat('count_occurrences', None) < min_code_occurrence_count
        ):
            invalid_entries.append(entry.to_token())
            continue
    
        # If we've made it here, then we want to keep this token
        valid_entries.append(entry)
    
    # Keep only the top `keep_n_max_occurrence_codes` tokens, sorted by occurrence count (if specified)
    if keep_n_max_occurrence_codes is not None:
        sorted_entries: List[TokenizerConfigEntry] = sorted(valid_entries, key=lambda x: x.get_stat('count_occurrences', None), reverse=True)
        valid_entries = sorted_entries[:keep_n_max_occurrence_codes]
        invalid_entries = sorted_entries[keep_n_max_occurrence_codes:]

    return valid_entries, invalid_entries

def is_metadata_equal(metadata1: Dict, metadata2: Dict) -> bool:
    """Return TRUE if `metadata1` EXACTLY EQUALS `metadata2`"""
    # Handle special case of paths that get rewritten
    metadata_paths: List[str] = [ 'path_to_femr_extract' ]
    metadata1 = { key: val if not key in metadata_paths else os.path.basename(val) for key, val in metadata1.items() }
    metadata2 = { key: val if not key in metadata_paths else os.path.basename(val) for key, val in metadata2.items() }
    
    # Ignore `is_already_run`
    metadata1.pop('is_already_run', None)
    metadata2.pop('is_already_run', None)

    is_match: bool = True
    for key, val in metadata1.items():
        if key not in metadata2:
            return False
        if metadata2[key] != val:
            return False
    for key, val in metadata2.items():
        if key not in metadata1:
            return False
        if metadata1[key] != val:
            return False
    return is_match

class BaseTokenizer(PreTrainedTokenizer):
    path_to_tokenizer_config: str
    
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        assert self.path_to_tokenizer_config is not None, f"ERROR - `self.path_to_tokenizer_config` must be set on `init()`"
        assert hasattr(self, 'metadata'), f"ERROR - `self.metadata` must be set on `init()`"
        # Convert DictConfig to a regular dictionary if necessary
        if isinstance(self.metadata, DictConfig):
            self.metadata = OmegaConf.to_container(self.metadata, resolve=True)
        assert isinstance(self.metadata, dict), f"ERROR - `self.metadata` must be a dict, but got {type(self.metadata)}"
        self.path_to_tokenizer_version_dir: str = self.get_path_to_tokenizer_version_dir() # trigger creation of version folder if it doesn't exist

    
    ########################################################
    # Tokenization helpers
    ########################################################
    def convert_events_to_tokens(self, events: List[Event], **kwargs) -> List[str]:
        """Provide default implementation where one Event => one token"""
        tokens: List[str] = []
        for e in events:
            token: Optional[str] = self.convert_event_to_token(e, **kwargs)
            if token is not None:
                tokens.append(token)
        return tokens
    
    def convert_events_to_tokenized_events(self, events: List[Event], **kwargs) -> List[Event]:
        """Returns all events that DO get mapped to tokens"""
        tokenized_events: List[Event] = []
        for e in events:
            token: Optional[str] = self.convert_event_to_token(e, **kwargs)
            if token is not None:
                tokenized_events.append(e)
        return tokenized_events
    
    def convert_events_to_non_tokenized_events(self, events: List[Event], **kwargs) -> List[Event]:
        """Returns all events that DO NOT get mapped to tokens"""
        non_tokenized_events: List[Event] = []
        for e in events:
            token: Optional[str] = self.convert_event_to_token(e, **kwargs)
            if token is None:
                non_tokenized_events.append(e)
        return non_tokenized_events
    
    def convert_event_to_token(self, e: Event, **kwargs) -> Optional[str]:
        raise NotImplementedError("Must implement `self.convert_event_to_token()` in child class")

    ########################################################
    # Caching / versioning
    ########################################################
    def save(self):
        """Dump vocab for this tokenizer version"""
        json.dump({
            'timestamp' : datetime.datetime.now().isoformat(),
            'metadata' : self.metadata,
            'vocab' : self.get_vocab(),
        }, open(os.path.join(self.path_to_tokenizer_version_dir, 'vocab.json'), 'w'), indent=2)
        save_tokenizer_config_to_path(os.path.join(self.path_to_tokenizer_version_dir, 'tokenizer_config_filtered.json'), self.tokenizer_config)

    def get_path_to_tokenizer_version_dir(self) -> str:
        """
            Example path: /path/to/dataset/versions/

            The tokenizer can have multiple versions depending on its `self.metadata`. 
            This method returns the path to the folder containing the exact version that matches `self.metadata`.
            If no folder exists for this version, then creates a new folder and return that.
        """
        path_to_tokenizer_dir: str = os.path.dirname(self.path_to_tokenizer_config)
        path_to_versions_dir: str = os.path.join(path_to_tokenizer_dir, 'versions/')
        os.makedirs(path_to_versions_dir, exist_ok=True)
        
        # Find folder corresponding to this version
        for f in os.listdir(path_to_versions_dir):
            if not os.path.isdir(os.path.join(path_to_versions_dir, f)):
                continue
            # Read metadata in `f``
            f_metadata = json.load(open(os.path.join(path_to_versions_dir, f, 'metadata.json'), 'r'))
            
            # If `self.metadata` of this tokenizer exactly matches `metadata.json` in this folder, then return it
            if is_metadata_equal(self.metadata, f_metadata):
                return os.path.join(path_to_versions_dir, f)
        
        # No matching folders found, so create a new one for this version
        path_to_new_folder: str = os.path.join(path_to_versions_dir, datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S'))
        os.makedirs(path_to_new_folder, exist_ok=True)
        json.dump(self.metadata, open(os.path.join(path_to_new_folder, 'metadata.json'), 'w'), indent=2)
        print(f"Creating new folder for this version of the tokenizer at `{path_to_new_folder}` with metadata={self.metadata}")
        return path_to_new_folder
    
    def get_path_to_dataset_dir(self, dataset: 'FEMRDataset') -> str:
        """
            Example path: /path/to/dataset/
        
            The tokenizer can have certain dataset-specific properties. 
            We store those in a datasets/ folder within the versions/ folder.
            We make sure that the dataset we retrieve matches the metadata of the argument `dataset`.
        """
        path_to_datasets_dir: str = os.path.join(self.path_to_tokenizer_version_dir, 'datasets/')
        os.makedirs(path_to_datasets_dir, exist_ok=True)
        
        # Find folder corresponding to this version's dataset
        for f in os.listdir(path_to_datasets_dir):
            if not os.path.isdir(os.path.join(path_to_datasets_dir, f)):
                continue
            # Read metadata in `f``
            f_metadata = json.load(open(os.path.join(path_to_datasets_dir, f, 'metadata.json'), 'r'))
            
            # If `self.metadata` of this dataset exactly matches `metadata.json` in this folder, then return it
            if is_metadata_equal(dataset.metadata, f_metadata):
                return os.path.join(path_to_datasets_dir, f)

        # No matching folders found, so create a new one for this version's dataset
        path_to_new_folder: str = os.path.join(path_to_datasets_dir, datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S'))
        os.makedirs(path_to_new_folder, exist_ok=True)
        json.dump(dataset.metadata, open(os.path.join(path_to_new_folder, 'metadata.json'), 'w'), indent=2)
        print(f"Creating new folder for this dataset of this version of the tokenizer at `{path_to_new_folder}` with metadata={dataset.metadata}")
        return path_to_new_folder

    def get_seq_length(self, args: Tuple['FEMRDataset', int, int]) -> List[Tuple[int, int]]:
        """Given a dataset and a range of indices, return the sequence length of each patient in that range"""
        from hf_ehr.data.datasets import FEMRDataset
        dataset_metadata, start_idx, end_idx = args # type is: FEMRDataset, int, int
        
        # remove extraneous keys so that we can init FEMRDataset() without errors
        for key in [ 'cls', 'tokenizer_metadata', 'max_length' ]:
            if key in dataset_metadata: del dataset_metadata[key]

        dataset = FEMRDataset(**dataset_metadata)
        results: List[Tuple[int, int]] = []
        for idx in range(start_idx, end_idx):
            events: List[Event] = dataset.__getitem__(idx)[1]
            results.append((idx, len(self.__call__(events)['input_ids'][0])))
        return results

    def get_seq_length_per_patient(self, dataset, n_procs: int = 5, is_force_refresh: bool = False) -> List[int]:
        """
            Fetch the sequence length of every patient in `dataset`, save to cache, and return the list of lengths.
            If cache exists, then load from cache (unless `is_force_refresh` is set to TRUE).
            If cache doesn't exist or `is_force_refresh` is TRUE, then calculate the lengths in parallel. 
                For CLMBRTokenizer, `n_procs=5`, this takes ~5 hrs for 2.5M patients (train dataset)
                For DescTokenizer, `n_procs=5`, this takes ~11 hrs for 2.5M patients (train dataset)
        """
        # Check if cache exists
        path_to_dataset_dir: str = self.get_path_to_dataset_dir(dataset)
        path_to_cache_file: str = os.path.join(path_to_dataset_dir, 'seq_length_per_patient.json')

        if not is_force_refresh:
            # If NOT force refresh, try to load from cache
            if os.path.exists(path_to_cache_file):
                print(f"Loading `seq_length_per_patient.json` from `{path_to_cache_file}` for split=`{dataset.split}`")
                data: TokenizerSeqLengthPerPatientCache = json.load(open(path_to_cache_file, 'r'))
                seq_lengths: List[int] = data['seq_lengths']
                if (
                    len(seq_lengths) == len(dataset)
                    and is_metadata_equal(self.metadata, data.get('tokenizer_metadata'))
                    and is_metadata_equal(dataset.metadata, data.get('dataset_metadata'))
                ):
                    return seq_lengths
                print(f"The # of `seq_lengths` in `{path_to_cache_file}` didn't match this dataset's length ({len(seq_lengths)} != {len(dataset)}) or the `metadata` differed for tokenizer ({self.metadata} != {data['tokenizer_metadata']}) or dataset ({dataset.metadata} != {data['dataset_metadata']}), so recreating `seq_length_per_patient.json` from scratch now...")
            else:
                print(f"No `seq_length_per_patient.json` found at `{path_to_cache_file}` for split=`{dataset.split}`. Generating `seq_length_per_patient.json` now...")

        # Calculate seq lengths in parallel
        if n_procs == 1:
            tasks: List[Tuple] = [(dataset.metadata, start, min(len(dataset), start + 1),) for start in range(0, len(dataset), 1) ]
            results: List[List[Tuple[int,int]]] = [ self.get_seq_length(t) for t in tqdm(tasks, total=len(tasks), desc=f"tokenizer.get_seq_length_per_patient() | n_procs={n_procs}") ]
        else:
            chunk_size: int = 5_000
            tasks: List[Tuple] = [(dataset.metadata, start, min(len(dataset), start + chunk_size),) for start in range(0, len(dataset), chunk_size) ]
            with multiprocessing.Pool(processes=n_procs) as pool:
                results: List[List[Tuple[int,int]]] = list(tqdm(pool.imap_unordered(self.get_seq_length, tasks), total=len(tasks), desc=f"tokenizer.get_seq_length_per_patient() | n_procs={n_procs}"))
        flattened_results: List[Tuple[int, int]] = [ x for sublist in results for x in sublist ] # type: ignore
        seq_lengths: List[int] = [ x[1] for x in sorted(flattened_results, key=lambda x: x[0]) ]
        json.dump({ 
            'timestamp' : datetime.datetime.now().isoformat(), 
            'tokenizer_metadata' : self.metadata, 
            'dataset_metadata' : dataset.metadata, 
            'seq_lengths' : seq_lengths 
        }, open(path_to_cache_file, 'w'), indent=2)
        return seq_lengths

class BaseCodeTokenizer(BaseTokenizer):

    def __init__(self) -> None:
        # Create vocab
        self.special_tokens = [ '[BOS]', '[EOS]', '[UNK]', '[SEP]', '[PAD]', '[CLS]', '[MASK]']
        self.vocab = self.special_tokens + self.non_special_tokens

        # Map tokens -> idxs
        self.token_2_idx: Dict[str, int] = { x: idx for idx, x in enumerate(self.vocab) }
        self.idx_2_token: Dict[int, str] = { idx: x for idx, x in enumerate(self.vocab) }

        # Create tokenizer
        super().__init__(
            bos_token='[BOS]',
            eos_token='[EOS]',
            unk_token='[UNK]',
            sep_token='[SEP]',
            pad_token='[PAD]',
            cls_token='[CLS]',
            mask_token='[MASK]',
        )
        self.add_tokens(self.non_special_tokens)


    
    def __call__(self, 
                 batch_of_events: Union[List[Event], List[List[Event]]],
                 is_truncation_random: bool = False,
                 seed: int = 1,
                 **kwargs) -> Dict[str, torch.Tensor]:
        """Tokenize a batch of patient timelines, where each timeline is a list of event codes.
            We add the ability to truncate seqs at random time points.
            
            Expects as input a list of Events

            NOTE: Must set `is_split_into_words=True` b/c we've already pre-tokenized our inputs (i.e. we're passing in a List of tokens, not a string)
        """
        if not isinstance(batch_of_events[0], list):
            # List[Event] => List[List[Event]]
            batch_of_events = [ batch_of_events ] # type: ignore
        
        # First, convert all Events => ProtoTokens
        batch: List[List[str]] = [ self.convert_events_to_tokens(x) for x in batch_of_events ]

        # Second, add special tokens (if applicable)
        if kwargs.get("add_special_tokens", False):
            batch = [ [ self.cls_token, self.bos_token ] + x + [ self.eos_token ] for x in batch ]

        # Third, tokenize the batch
        if is_truncation_random:
            max_length: int = kwargs.get("max_length")
            if not max_length:
                raise ValueError(f"If you specify `is_truncation_random`, then you must also provide a non-None value for `max_length`")

            # Tokenize without truncation
            kwargs.pop('max_length')
            kwargs.pop('truncation')
            tokenized_batch: Dict[str, torch.Tensor] = super().__call__(batch, **kwargs, truncation=None, is_split_into_words=True)

            # Truncate at random positions
            random.seed(seed)
            start_idxs: List[int] = []
            for timeline in tokenized_batch['input_ids']:
                length: int = (timeline != self.pad_token_id).sum() # count of non-PAD tokens
                if length > max_length:
                    # Calculate a random start index
                    start_index: int = random.randint(0, length - max_length)
                    start_idxs.append(start_index)
                else:
                    start_idxs.append(0)
                    
            for key in tokenized_batch.keys():
                truncated_batch: List[List[int]] = []
                for idx, timeline in enumerate(tokenized_batch[key]):
                    new_timeline = timeline[start_idxs[idx]:start_idxs[idx] + max_length]
                    assert new_timeline.shape[0] <= max_length, f"Error in truncating by random positions: new_timeline.shape = {new_timeline.shape[0]} !<= max_length={max_length}"
                    truncated_batch.append(new_timeline)
                if kwargs.get('return_tensors') == 'pt':
                    tokenized_batch[key] = torch.stack(truncated_batch, dim=0)
                else:
                    tokenized_batch[key] = truncated_batch
        else:
            try:
                tokenized_batch: Dict[str, torch.Tensor] = super().__call__(batch, **kwargs, is_split_into_words=True)
            except Exception as e:
                print("Exception in BaseCodeTokenizer.__call__():", e)
                raise e

        return tokenized_batch

    """Mandatory overwrites of base class"""
    @property
    def vocab_size(self) -> int:
        return len(self.vocab)

    def get_vocab(self) -> Dict[str, int]:
        return self.token_2_idx
    
    def tokenize(self, text: str, **kwargs) -> int:
        """Here, `text` will be a single code (e.g. "LOINC/13"), so just map it back to itself"""
        return [ text ]

    def _tokenize(self, text: str, **kwargs) -> int:
        """Here, `text` will be a single code (e.g. "LOINC/13"), so just map it back to itself"""
        return [ text ]

    def _convert_token_to_id(self, token: str) -> int:
        return self.token_2_idx[token]

    def _convert_id_to_token(self, index: int) -> str:
        raise self.idx_2_token[index]

class CookbookTokenizer(BaseCodeTokenizer):
    """
        Settings:
            is_remap_numerical_codes_to_quantiles: bool
                - If TRUE, then remap numericals to buckets based on quantile of value
            excluded_vocabs: Optional[List[str]]
                - List of vocabs to exclude from the tokenizer. Determined by the first part of the code before the '/' (e.g. "OBS" in "OBS/1234")
            min_code_occurrence_count: Optional[int]
                - Only keep tokens with >= `min_code_occurrence_count` total occurrences in our dataset
    """
    def __init__(self, 
                 path_to_tokenizer_config: str, 
                 metadata: Optional[Dict[str, Any]] = None) -> None:
        self.path_to_tokenizer_config: str = path_to_tokenizer_config
        self.tokenizer_config: List[TokenizerConfigEntry] = load_tokenizer_config_from_path(path_to_tokenizer_config)
        
        # Set metadata
        self.metadata: Dict[str, Any] = {} if metadata is None else dict(metadata)
        self.metadata['cls'] = 'CookbookTokenizer'

        # Fetches ATT metadata - By default, it'll be done in CEHR-GPT style
        self.is_add_visit_start: bool = metadata.get('is_add_visit_start', True)
        self.is_add_visit_end: bool = metadata.get('is_add_visit_end', True)
        self.is_add_day_att: bool = metadata.get('is_add_day_att', True)
        self.is_add_day_week_month_att: bool = metadata.get('is_add_day_week_month_att', False)

        # Pop the flags from metadata after initializing them to avoid recreating the config
        for key in ['is_add_visit_start', 'is_add_visit_end', 'is_add_day_att', 'is_add_day_week_month_att']:
            if key in self.metadata:
                self.metadata.pop(key)

        # Metadata
        self.is_remap_numerical_codes_to_quantiles: bool = metadata.get('is_remap_numerical_codes_to_quantiles', False)
        self.excluded_vocabs: Optional[Set[str]] = { x.lower() for x in metadata.get('excluded_vocabs', {}) } if metadata.get('excluded_vocabs', {}) else None # type: ignore
        self.min_code_occurrence_count: Optional[int] = metadata.get('min_code_occurrence_count', None)
        self.keep_n_max_occurrence_codes: Optional[int] = metadata.get('keep_n_max_occurrence_codes', None)

        # Apply filtering
        self.tokenizer_config, self.excluded_tokens = filter_tokenizer_config(self.tokenizer_config, 
                                                                              self.excluded_vocabs, 
                                                                              self.min_code_occurrence_count,
                                                                              self.keep_n_max_occurrence_codes)
        # Tokens
        self.code_2_token = {} # [key] = token; [val] = { 'type' : str, 'tokenization' : dict, 'token' : str }
        self.non_special_tokens: List[str] = []
        
        # Preprocess tokenizer config for quick access
        for entry in self.tokenizer_config: # NOTE: Takes ~10 seconds for 1.5M tokens
            if entry.code not in self.code_2_token: self.code_2_token[entry.code] = {}
            if entry.type not in self.code_2_token[entry.code]: self.code_2_token[entry.code][entry.type] = []
            self.code_2_token[entry.code][entry.type].append({
                'tokenization': entry.tokenization,
                'token' : entry.to_token(),
            })
            self.non_special_tokens.append(entry.to_token())
        
        # Create tokenizer
        super().__init__()

    def convert_event_to_token(self, e: Event, **kwargs) -> Optional[str]:
        """NOTE: This is basically the same as the CLMBR tokenizer's version."""
        event_code = e.code
        # If code isn't in vocab => ignore
        if event_code not in self.code_2_token:
            return None
        
        event_value = e.value
        # If numerical code...
        if (
            'numerical_range' in self.code_2_token[event_code] # `numerical_range` is a valid type for this code
            and event_value is not None # `value` is not None
            and ( # `value` is numeric
                isinstance(event_value, float)
                or isinstance(event_value, int)
            )
        ):
            for token_range in self.code_2_token[event_code]['numerical_range']:
                assert 'token' in token_range, f"ERROR - Missing 'token' for code={e.code},type=numerical_range in self.code_2_token: {self.code_2_token[e.code]['numerical_range']}"
                assert 'tokenization' in token_range, f"ERROR - Missing 'tokenization' for code={e.code},type=numerical_range in self.code_2_token: {self.code_2_token[e.code]['numerical_range']}"
                token: str = token_range['token']
                unit: str = token_range['tokenization']['unit']
                range_start: float = token_range['tokenization']['range_start']
                range_end: float = token_range['tokenization']['range_end']
                if range_start <= event_value <= range_end and e.unit == unit:
                    return token
            return None

        # If textual code...
        if (
            'categorical' in self.code_2_token[event_code] # `categorical` is a valid type for this code
            and event_value is not None # `value` is not None
            and event_value != '' # `value` is not blank
            and ( # `value` is textual
                isinstance(event_value, str)
            )
        ):
            for categorical_value in self.code_2_token[event_code]['categorical']:
                assert 'token' in categorical_value, f"ERROR - Missing 'token' for code={event_code},type=categorical in self.code_2_token: {self.code_2_token[event_code]['categorical']}"
                assert 'tokenization' in categorical_value, f"ERROR - Missing 'tokenization' for code={event_code},type=categorical in self.code_2_token: {self.code_2_token[event_code]['categorical']}"
                if event_value in categorical_value['tokenization']['categories']:
                    token: str = categorical_value['token']
                    return token
            return None

        # If just vanilla code...
        if (
            'code' in self.code_2_token[event_code] # `code` is a valid type for this code
        ):
            token: str = self.code_2_token[event_code]['code'][0]['token']
            return token

        return None
    
    def convert_events_to_tokens(self, events: List[Event], **kwargs) -> List[str]:
        """Converts a list of events into a list of tokens"""
        tokens: List[str] = []
        current_visit_end: Optional[datetime.datetime] = None # track the end time of the currently active visit
        previous_visit_end: Optional[datetime.datetime] = None # track the end time of the immediately preceding visit

        for e in events:

            # Check if we need to add a visit end token
            if current_visit_end is not None and (
                e.start > current_visit_end # If we have [VISIT A = { TOKEN 1, TOKEN 2 }] [TOKEN 3], then add a visit end token before [TOKEN 3]
                or "Visit" in e.code # If we have [VISIT A = { TOKEN 1, TOKEN 2 }] [VISIT B = { TOKEN 3, TOKEN 4 }], then add a visit end token after [VISIT A]
            ):
                # This token occurs after the currently active visit ends, so end it (if exists)
                if self.is_add_visit_end:
                    tokens.append(self.visit_end)
                current_visit_end = None

            # Check if the event is a visit
            if "Visit" in e.code:
                    
                # Add ATT Tokens, if applicable
                # This will be inserted between the prior visit and the current visit
                if previous_visit_end is not None:
                    interval: float = (e.start - previous_visit_end).days # Time (in days) between this visit's start and the immediately preceding visit's end
                    assert interval >= 0, f"Interval has value = {interval} but should always be positive, but fails on {e}."
                    
                    if self.is_add_day_att:
                        if interval <= 1080:
                            att = self.day_atts_cehr_gpt[interval - 1]
                        else:
                            att = self.long_att_cehr_gpt
                        tokens.append(att)
                    elif self.is_add_day_week_month_att:
                        if interval < 7:
                            att = self.day_atts_cehr_bert[interval - 1]
                        elif 7 <= interval < 30:
                            att = self.week_atts[(interval // 7) - 1]
                        elif 30 <= interval < 360:
                            att = self.month_atts[(interval // 30) - 1]
                        else:
                            att = self.long_att_cehr_bert
                        tokens.append(att)

                # Add visit start token, if applicable
                # if self.is_add_visit_start:
                #     tokens.append(self.visit_start)
            
                # Add token itself
                token = self.convert_event_to_token(e, **kwargs)
                if token:
                    tokens.append(token)

                # Keep track of this visit's end
                current_visit_end = e.end
                previous_visit_end = e.end
            else:
                token = self.convert_event_to_token(e, **kwargs)
                if token:
                    tokens.append(token)
        
        return tokens

class CLMBRTokenizer(BaseCodeTokenizer):
    def __init__(self, path_to_tokenizer_config: str) -> None:
        self.path_to_tokenizer_config: str = path_to_tokenizer_config
        self.tokenizer_config: List[TokenizerConfigEntry] = load_tokenizer_config_from_path(path_to_tokenizer_config)
        
        # Set metadata        
        self.metadata: Dict[str, Any] = {}
        self.metadata['cls'] = 'CLMBRTokenizer'

        # Preprocess tokenizer config for quick access
        self.code_2_token = {} # [key] = token; [val] = { 'type' : str, 'tokenization' : dict, 'token' : str }
        self.non_special_tokens: List[str] = []
        for entry in self.tokenizer_config:
            if entry.code not in self.code_2_token: self.code_2_token[entry.code] = {}
            if entry.type not in self.code_2_token[entry.code]: self.code_2_token[entry.code][entry.type] = []
            self.code_2_token[entry.code][entry.type].append({
                'tokenization': entry.tokenization,
                'token' : entry.to_token(),
            })
            self.non_special_tokens.append(entry.to_token())

        # assert len(self.non_special_tokens) == 39811, f"ERROR - Expected 39811 self.non_special_tokens, but got {len(self.non_special_tokens)}"

        # Create tokenizer
        super().__init__()

    def convert_event_to_token(self, e: Event, **kwargs) -> Optional[str]:
        # If code isn't in CLMBR vocab => ignore
        if e.code not in self.code_2_token:
            return None
        
        # If numerical code...
        if (
            'numerical_range' in self.code_2_token[e.code] # `numerical_range` is a valid type for this code
            and e.value is not None # `value` is not None
            and ( # `value` is numeric
                isinstance(e.value, float)
                or isinstance(e.value, int)
            )
        ):
            # NOTE: CLMBR ignores units
            for token_range in self.code_2_token[e.code]['numerical_range']:
                assert 'token' in token_range, f"ERROR - Missing 'token' for code={e.code},type=numerical_range in self.code_2_token: {self.code_2_token[e.code]['numerical_range']}"
                assert 'tokenization' in token_range, f"ERROR - Missing 'tokenization' for code={e.code},type=numerical_range in self.code_2_token: {self.code_2_token[e.code]['numerical_range']}"
                token: str = token_range['token']
                range_start: float = token_range['tokenization']['range_start']
                range_end: float = token_range['tokenization']['range_end']
                if range_start <= e.value <= range_end:
                    return token
            return None

        # If textual code...
        if (
            'categorical' in self.code_2_token[e.code] # `categorical` is a valid type for this code
            and e.value is not None # `value` is not None
            and e.value != '' # `value` is not blank
            and ( # `value` is textual
                isinstance(e.value, str)
            )
        ):
            for categorical_value in self.code_2_token[e.code]['categorical']:
                assert 'token' in categorical_value, f"ERROR - Missing 'token' for code={e.code},type=categorical in self.code_2_token: {self.code_2_token[e.code]['categorical']}"
                assert 'tokenization' in categorical_value, f"ERROR - Missing 'tokenization' for code={e.code},type=categorical in self.code_2_token: {self.code_2_token[e.code]['categorical']}"
                if e.value in categorical_value['tokenization']['categories']:
                    token: str = categorical_value['token']
                    return token
            return None

        # If just vanilla code...
        if (
            'code' in self.code_2_token[e.code] # `code` is a valid type for this code
        ):
            token: str = self.code_2_token[e.code]['code'][0]['token']
            return token

        return None


class CEHRTokenizer(BaseCodeTokenizer):
    def __init__(self, path_to_tokenizer_config: str, metadata: Optional[Dict[str, Any]] = None) -> None:
        self.path_to_tokenizer_config: str = path_to_tokenizer_config
        self.tokenizer_config: List[TokenizerConfigEntry] = load_tokenizer_config_from_path(path_to_tokenizer_config)
        
        # Set metadata        
        self.metadata: Dict[str, Any] = {} if metadata is None else dict(metadata)
        self.metadata['cls'] = 'CEHRTokenizer'

        # Fetches ATT metadata - By default, it'll be done in CEHR-GPT style
        self.is_add_visit_start: bool = self.metadata.get('is_add_visit_start', True)
        self.is_add_visit_end: bool = self.metadata.get('is_add_visit_end', True)
        self.is_add_day_att: bool = self.metadata.get('is_add_day_att', True)
        self.is_add_day_week_month_att: bool = self.metadata.get('is_add_day_week_month_att', False)

        # Initialize special tokens for visits and time intervals (ATT tokens)
        self.visit_start = "[VISIT START]"
        self.visit_end = "[VISIT END]"
        self.day_atts_cehr_gpt = [f"[DAY {i}]" for i in range(1, 1081)]  # Days 1 to 1080
        self.long_att_cehr_gpt = "[LONG TERM]"
        self.day_atts_cehr_bert = [f"[DAY {i}]" for i in range(1, 7)]  # Days 1 to 6
        self.week_atts = [f"[WEEK {i}]" for i in range(1, 5)]  # Weeks 1 to 4
        self.month_atts = [f"[MONTH {i}]" for i in range(1, 13)]  # Months 1 to 12
        self.long_att_cehr_bert = "[LONG TERM]"

        # Add special tokens to the tokenizer
        self.special_tokens = [
            self.visit_start, self.visit_end, self.long_att_cehr_gpt, self.long_att_cehr_bert
        ] + self.day_atts_cehr_gpt + self.day_atts_cehr_bert + self.week_atts + self.month_atts

        # Preprocess tokenizer config for quick access
        self.code_2_token = {} # [key] = token; [val] = { 'type' : str, 'tokenization' : dict, 'token' : str }
        self.non_special_tokens: List[str] = []
        for entry in self.tokenizer_config:
            if entry.code not in self.code_2_token: self.code_2_token[entry.code] = {}
            if entry.type not in self.code_2_token[entry.code]: self.code_2_token[entry.code][entry.type] = []
            self.code_2_token[entry.code][entry.type].append({
                'tokenization': entry.tokenization,
                'token' : entry.to_token(),
            })
            self.non_special_tokens.append(entry.to_token())

        # assert len(self.non_special_tokens) == 39811, f"ERROR - Expected 39811 self.non_special_tokens, but got {len(self.non_special_tokens)}"

        # Create tokenizer
        super().__init__()

    def convert_event_to_token(self, e: Event, **kwargs) -> Optional[str]:
        # If code isn't in CLMBR vocab => ignore
        if e.code not in self.code_2_token:
            return None
        
        # If numerical code...
        if (
            'numerical_range' in self.code_2_token[e.code] # `numerical_range` is a valid type for this code
            and e.value is not None # `value` is not None
            and ( # `value` is numeric
                isinstance(e.value, float)
                or isinstance(e.value, int)
            )
        ):
            # NOTE: CLMBR ignores units
            for token_range in self.code_2_token[e.code]['numerical_range']:
                assert 'token' in token_range, f"ERROR - Missing 'token' for code={e.code},type=numerical_range in self.code_2_token: {self.code_2_token[e.code]['numerical_range']}"
                assert 'tokenization' in token_range, f"ERROR - Missing 'tokenization' for code={e.code},type=numerical_range in self.code_2_token: {self.code_2_token[e.code]['numerical_range']}"
                token: str = token_range['token']
                range_start: float = token_range['tokenization']['range_start']
                range_end: float = token_range['tokenization']['range_end']
                if range_start <= e.value <= range_end:
                    return token
            return None

        # If textual code...
        if (
            'categorical' in self.code_2_token[e.code] # `categorical` is a valid type for this code
            and e.value is not None # `value` is not None
            and e.value != '' # `value` is not blank
            and ( # `value` is textual
                isinstance(e.value, str)
            )
        ):
            for categorical_value in self.code_2_token[e.code]['categorical']:
                assert 'token' in categorical_value, f"ERROR - Missing 'token' for code={e.code},type=categorical in self.code_2_token: {self.code_2_token[e.code]['categorical']}"
                assert 'tokenization' in categorical_value, f"ERROR - Missing 'tokenization' for code={e.code},type=categorical in self.code_2_token: {self.code_2_token[e.code]['categorical']}"
                if e.value in categorical_value['tokenization']['categories']:
                    token: str = categorical_value['token']
                    return token
            return None

        # If just vanilla code...
        if (
            'code' in self.code_2_token[e.code] # `code` is a valid type for this code
        ):
            token: str = self.code_2_token[e.code]['code'][0]['token']
            return token

        return None
    
    def convert_events_to_tokens(self, events: List[Event]) -> List[str]:
        """
        Convert a list of events into a list of tokens, inserting ATT tokens based on time intervals between visits.
        """
        tokens = []
        current_visit_end = None
        previous_visit_end = None

        for e in events:
            # Add visit end token if the event is after the previous visit's end
            if current_visit_end is not None and e.start > current_visit_end:
                if self.is_add_visit_end:
                    tokens.append(self.visit_end)
                current_visit_end = None

            # Handling visits and adding ATT tokens
            if "Visit" in e.code:
                # Add ATT tokens between visits based on time intervals
                if previous_visit_end is not None:
                    interval = (e.start - previous_visit_end).days
                    if interval >= 0:
                        if self.is_add_day_att:
                            if interval <= 1080:
                                tokens.append(self.day_atts_cehr_gpt[interval - 1])
                            else:
                                tokens.append(self.long_att_cehr_gpt)
                        elif self.is_add_day_week_month_att:
                            if interval < 7:
                                tokens.append(self.day_atts_cehr_bert[interval - 1])
                            elif 7 <= interval < 30:
                                tokens.append(self.week_atts[(interval // 7) - 1])
                            elif 30 <= interval < 360:
                                tokens.append(self.month_atts[(interval // 30) - 1])
                            else:
                                tokens.append(self.long_att_cehr_bert)

                # Add visit start token
                if self.is_add_visit_start:
                    tokens.append(self.visit_start)

                # Convert visit event to token and add it
                token = self.convert_event_to_token(e)
                if token:
                    tokens.append(token)

                current_visit_end = e.end
                previous_visit_end = e.end
            else:
                # Convert non-visit events to tokens
                token = self.convert_event_to_token(e)
                if token:
                    tokens.append(token)

        return tokens

class DescTokenizer(BaseTokenizer):
    """Converts codes => textual descriptions, then tokenizes using a normal text tokenizer (e.g. BERT)
    """
    def __init__(self, 
                 path_to_tokenizer_config: str, 
                 metadata: Optional[Dict[str, Any]] = None) -> None:
        assert metadata is not None, f"ERROR - `metadata` must be provided, but got {metadata}"
        assert 'desc_emb_tokenizer' in metadata, f"ERROR - `metadata` must contain a 'desc_emb_tokenizer' key, but got {metadata}"
        self.path_to_tokenizer_config: str = path_to_tokenizer_config
        self.tokenizer_config: List[TokenizerConfigEntry] = load_tokenizer_config_from_path(path_to_tokenizer_config)
        
        # Set metadata
        self.metadata: Dict[str, Any] = {} if metadata is None else dict(metadata)
        self.metadata['cls'] = 'DescTokenizer'

        # Load underlying textual tokenizer
        self.desc_emb_tokenizer: str = metadata['desc_emb_tokenizer']
        self.tokenizer = AutoTokenizer.from_pretrained(self.desc_emb_tokenizer)
        self.event_separator: str = ' ' # character that gets inserted between Events when transformed into their textual descriptions
        
        # Metadata
        metadata = {} if metadata is None else metadata
        self.excluded_vocabs: Optional[Set[str]] = { x.lower() for x in metadata.get('excluded_vocabs', {}) } if metadata.get('excluded_vocabs', {}) else None # type: ignore
        self.min_code_occurrence_count: Optional[int] = metadata.get('min_code_occurrence_count', None)
        self.keep_n_max_occurrence_codes: Optional[int] = metadata.get('keep_n_max_occurrence_codes', None)
        
        # Apply filtering
        self.tokenizer_config, self.excluded_tokens = filter_tokenizer_config(self.tokenizer_config, 
                                                                              self.excluded_vocabs, 
                                                                              self.min_code_occurrence_count,
                                                                              self.keep_n_max_occurrence_codes)
        
        # Preprocess tokenizer config for quick access
        self.code_2_desc: Dict[str, str] = {}
        # initialize non special tokens list
        self.non_special_tokens: List[str] = []
        for entry in self.tokenizer_config:
            if entry.description is not None:
                self.code_2_desc[entry.code] = entry.description
                self.non_special_tokens.append(entry.description)

        # Define special tokens 
        self.special_tokens: List[str] = [
            self.tokenizer.bos_token,
            self.tokenizer.eos_token,
            self.tokenizer.unk_token,
            self.tokenizer.sep_token,
            self.tokenizer.pad_token,
            self.tokenizer.cls_token,
            self.tokenizer.mask_token,
        ]
        
        super().__init__(
            bos_token=self.tokenizer.bos_token,
            eos_token=self.tokenizer.eos_token,
            unk_token=self.tokenizer.unk_token,
            sep_token=self.tokenizer.sep_token,
            pad_token=self.tokenizer.pad_token,
            cls_token=self.tokenizer.cls_token,
            mask_token=self.tokenizer.mask_token
        )
    
    def convert_event_to_token(self, e: Event, **kwargs) -> Optional[str]:
        """Converts a single event to its token representation"""
        if e.code not in self.code_2_desc:
            return None
        return self.code_2_desc[e.code]

    def __call__(self, 
                 batch_of_events: Union[List[Event], List[List[Event]]],
                 is_truncation_random: bool = False,
                 seed: int = 1,
                 **kwargs) -> Dict[str, torch.Tensor]:
        """
            Tokenize a batch of patient timelines, where each timeline is a list of event codes.
            We add the ability to truncate seqs at random time points.
            
            Expects as input a list of Events
        """
        if not isinstance(batch_of_events[0], list):
            # List[Event] => List[List[Event]]
            batch_of_events = [ batch_of_events ] # type: ignore
        
        # First, convert all Events => ProtoTokens
        batch: List[List[str]] = [ self.convert_events_to_tokens(x) for x in batch_of_events ] 
        
        # Second, add special tokens (if applicable)
        if kwargs.get("add_special_tokens", False):
            batch = [ [ self.cls_token, self.bos_token ] + x + [ self.eos_token ] for x in batch ]

        # Concatenate all strings together for tokenization by traditional HF tokenizer
        # List[List[str]] => List[str]
        #batch = [ self.event_separator.join(x) for x in batch ] # type: ignore
        # Ensure proper filtering of None values
        batch = [self.event_separator.join(filter(None, x)) for x in batch]

        if is_truncation_random:
            max_length: int = kwargs.get("max_length")
            if not max_length:
                raise ValueError(f"If you specify `is_truncation_random`, then you must also provide a non-None value for `max_length`")

            # Tokenize without truncation
            if 'max_length' in kwargs:
                del kwargs['max_length']
            if 'truncation' in kwargs:
                del kwargs['truncation']
            tokenized_batch: Dict[str, torch.Tensor] = self.tokenizer.__call__(batch, **kwargs, truncation=None)

            # Truncate at random positions
            random.seed(seed)
            start_idxs: List[int] = []
            for timeline in tokenized_batch['input_ids']:
                length: int = (timeline != self.pad_token_id).sum() # count of non-PAD tokens
                if length > max_length:
                    # Calculate a random start index
                    start_index: int = random.randint(0, length - max_length)
                    start_idxs.append(start_index)
                else:
                    start_idxs.append(0)
                    
            for key in tokenized_batch.keys():
                truncated_batch: List[List[int]] = []
                for idx, timeline in enumerate(tokenized_batch[key]):
                    new_timeline = timeline[start_idxs[idx]:start_idxs[idx] + max_length]
                    assert new_timeline.shape[0] <= max_length, f"Error in truncating by random positions: new_timeline.shape = {new_timeline.shape[0]} !<= max_length={max_length}"
                    truncated_batch.append(new_timeline)
                if kwargs.get('return_tensors') == 'pt':
                    tokenized_batch[key] = torch.stack(truncated_batch, dim=0)
                else:
                    tokenized_batch[key] = truncated_batch
        else:
            tokenized_batch: Dict[str, torch.Tensor] = self.tokenizer.__call__(batch, **kwargs)

        return tokenized_batch

    """Mandatory overwrites of base class"""
    @property
    def vocab_size(self) -> int:
        return len(self.tokenizer.get_vocab())

    def get_vocab(self) -> Dict[str, int]:
        return self.tokenizer.get_vocab()

    def _tokenize(self, text: str, **kwargs):
        return self.tokenizer._tokenize(text)

    def _convert_token_to_id(self, token: str) -> int:
        return self.tokenizer._convert_token_to_id(token)

    def _convert_id_to_token(self, index: int) -> str:
        return self.tokenizer._convert_id_to_token(index)

def torch_mask_tokens(tokenizer: BaseTokenizer, 
                      inputs: Any, 
                      mlm_prob: float, 
                      special_tokens_mask: Optional[Any] = None) -> Tuple[Any, Any]:
    """
        Prepare masked tokens inputs/labels for masked language modeling: 80% MASK, 10% random, 10% original.
        
        Taken from: https://github.com/huggingface/transformers/blob/09f9f566de83eef1f13ee83b5a1bbeebde5c80c1/src/transformers/data/data_collator.py#L782
    """
    labels = inputs.clone()
    # We sample a few tokens in each sequence for MLM training (with probability `mlm_prob`)
    probability_matrix = torch.full(labels.shape, mlm_prob)
    if special_tokens_mask is None:
        special_tokens_mask = [
            tokenizer.get_special_tokens_mask(val, already_has_special_tokens=True) for val in labels.tolist()
        ]
        special_tokens_mask = torch.tensor(special_tokens_mask, dtype=torch.bool)
    else:
        special_tokens_mask = special_tokens_mask.bool()

    probability_matrix.masked_fill_(special_tokens_mask, value=0.0)
    masked_indices = torch.bernoulli(probability_matrix).bool()
    labels[~masked_indices] = -100  # We only compute loss on masked tokens
    
    # Check if mask_token is set properly
    if tokenizer.mask_token is None:
        raise ValueError("The tokenizer's mask_token is not set.")

    # 80% of the time, we replace masked input tokens with tokenizer.mask_token ([MASK])
    indices_replaced = torch.bernoulli(torch.full(labels.shape, 0.8)).bool() & masked_indices
    mask_token_id = tokenizer.convert_tokens_to_ids(tokenizer.mask_token)
    if mask_token_id is None:
        raise ValueError(f"The mask token {tokenizer.mask_token} could not be converted to an ID.")
    inputs[indices_replaced] = mask_token_id

    # 10% of the time, we replace masked input tokens with random word
    indices_random = torch.bernoulli(torch.full(labels.shape, 0.5)).bool() & masked_indices & ~indices_replaced
    random_words = torch.randint(len(tokenizer), labels.shape, dtype=torch.long)
    inputs[indices_random] = random_words[indices_random]

    # The rest of the time (10% of the time) we keep the masked input tokens unchanged
    return inputs, labels

def collate_femr_timelines(batch: List[Tuple[int, List[Event]]],
                             tokenizer: BaseTokenizer, 
                             dataset_name: str, # 'FEMRDataset' or 'AllTokensFEMRDataset'
                             max_length: int,
                             is_truncation_random: bool = False,
                             is_mlm: bool = False,
                             mlm_prob: float = 0.15,
                             seed: int = 1) -> Dict[str, Any]:
    """
        Collate function for FEMR timelines
        Truncate or pad to max length in batch.
    """
    timelines: List[List[Event]] = [ x[1] for x in batch if len(x[1]) > 0 ] # remove empty timelines
    if dataset_name == 'AllTokensFEMRDataset':
        # For AllTokensFEMRDataset, we are given explicit (start, end) idx's to subselect from each patient's timeline
        tokens: Dict[str, Float[torch.Tensor, 'B max_length']] = tokenizer(timelines, 
                                                                            truncation=True, 
                                                                            padding=True,
                                                                            max_length=max_length,
                                                                            is_truncation_random=False,
                                                                            add_special_tokens=True,
                                                                            seed=seed, 
                                                                            return_tensors='pt')
    elif dataset_name == 'FEMRDataset':
        # For FEMRDataset, truncate timeline per usual
        tokens: Dict[str, Float[torch.Tensor, 'B max_length']] = tokenizer(timelines, 
                                                                            truncation=True, 
                                                                            padding=True, 
                                                                            max_length=max_length,
                                                                            is_truncation_random=is_truncation_random,
                                                                            add_special_tokens=True,
                                                                            seed=seed, 
                                                                            return_tensors='pt')
    else:
        raise ValueError(f"ERROR - Unsupported 'dataset_name' of: `{dataset_name}`")
    
    # Set labels
    if is_mlm:
        # Masked LM
        tokens["input_ids"], tokens["labels"] = torch_mask_tokens(tokenizer, tokens["input_ids"], mlm_prob)
    else:
        # Causal LM
        tokens['labels'] = tokens['input_ids']

    return {
        'patient_ids' : [ x[0] for x in batch ],
        'tokens' : tokens,
    }

if __name__ == '__main__':
    # Load v8 dataset
    print("Loading v8 dataset...")
    path_to_femr_extract: str = PATH_TO_FEMR_EXTRACT_v8
    dataset = FEMRDataset(path_to_femr_extract, split='train', is_debug=False)
    
    
    dataset = FEMRDataset(path_to_femr_extract, split='test', is_debug=False)
    tokenizer = DescTokenizer(PATH_TO_TOKENIZER_DESC_v8_CONFIG,
                              metadata={ 
                                'desc_emb_tokenizer' : 'bert-base-uncased', 
                                'excluded_vocabs' : ['OBS'] 
                              })
    tokenizer.get_seq_length_per_patient(dataset, n_procs=5)
    